//package org.seamcat.plugins;

import org.seamcat.model.propagation.*;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

/**
 * 
 * @author Karl Koch, karl.koch@bnetza.de
 * @version <strong>For presentation to the SEAMCAT workshop June 2013</strong>
 */
public class RandomDistanceWorkshop implements PropagationPluginModel {

	private final int URBAN = 5, SUBURBAN = 3, RURAL = 1; // values directly used as factors
	private int genEnv;
	private boolean envShouldBeConsidered;
	private Double maxRandomFactor;
	private Double stddev;
	private Random random = new Random();

	/**
	 * called through the interface by each link which uses this plugin as propagation model
	 * it returns the calculated path loss
	 */
	@Override	
	public double evaluate(double frequency, double distance, double TxHeight, double RxHeight, boolean variation,
			Parameter... params) {

		double loss = 0, randomFactor = 1.0; // initializes local

		initPlugin(params);

		loss = Math.sqrt(distance); // the main path loss calculation

		if (randomFactor > 0)
			loss *= (maxRandomFactor * Math.random() + 1); // avoids a path loss of zero

		if (envShouldBeConsidered)
			loss *= genEnv; // considers an additional factor if enabled 

		if (variation)
			loss += stddev * random.nextGaussian();

		return loss;
	}

	/**
	 * reads the parameter to be used by the PP
	 * 
	 * @param params
	 */
	private void initPlugin(Parameter[] params) { 
		String env = params[0].getValue();
		if (env.contentEquals("URBAN"))
			genEnv = URBAN;
		else if (env.endsWith("RURAL"))
			genEnv = RURAL;
		else
			genEnv = SUBURBAN;

		envShouldBeConsidered = params[1].isEnabled();

		maxRandomFactor = params[2].getValue();

		stddev = params[3].getValue();

	}

	@Override
	public List<ParameterDefinition> getParameterDefinitions() {
		List<ParameterDefinition> parameters = new ArrayList<ParameterDefinition>();

		parameters.add(new SelectionParameter("General Environment", "URBAN", "RURAL", "SUBURBAN"));

		parameters.add(new BooleanParameter("consider General Environment").defaultValue(false).description(
				"if selected an additional factor of 1 (rural), 3 (suburban) or 5 (urban) is considered "));

		parameters.add(new DoubleParameter("max value for random factor").defaultValue(10.0).unit("")
				.description("up to this the sqrt of the distance is manipulated"));

		parameters.add(new DoubleParameter("Standard deviation").defaultValue(3.5).unit("dB")
				.description("used if variation is selected"));
		
		Double defaultValue = null;
		parameters.add(new DoubleParameter("name").defaultValue(defaultValue).unit("unit").description("description").optional());

		return parameters;
	}

}
