//package plugins;

/*
 * This tool has been originally created by Christian Peters <Ciber DK>
 * 
 * modified by 
 * Karl Koch
 * BNetzA 412-10
 * fon: +49 (0) 681 9330 515
 * email: karl.koch@bnetza.de; karl.koch@ties.itu.int
 *
 * 
 */

import java.util.List;
import java.util.Vector;

import org.seamcat.mathematics.Mathematics;
import org.seamcat.model.plugin.*;
import org.seamcat.model.technical.exception.ModelException;
import org.seamcat.propagation.FreeSpaceModel;

public class SensingForWorkshop implements PostProcessingPlugin, ConsistencyCheck {

	private PluginDistribution defaultDistribution;
	private boolean checkPropagationModel;
	private Double threshold = -90.0;
	private Vector<DataCollectedForWorkshop> devicesActive;
	private Vector<DataCollectedForWorkshop> devicesInactive;
	private DataCollectedForWorkshop device;
	private double transmitPower;
	private Function2D antennaPattern;
	private FreeSpaceModel localFreeSpace;
	private double frequencyLBT;

	public void init(ParameterFactory param) {
		defaultDistribution = param.createDistribution(PluginDistribution.DistributionType.DiscreteUniform, 0, 360);

	}

	public void cleanUp() {
		// TODO Auto-generated method stub
	}

	public String getDescription() {
		return "This plug-in is for presentation to the SEAMCAT workshop June 2013";
	}

	/**
	 * the main method called through the PostProcessingPluginWrapper by the EGE
	 * after each snapshot
	 */
	public void process(ScenarioInfo scenario) {
		if (scenario.getCurrentEventNumber() < 2) {
			initPlugin(scenario);
		} else if (scenario.getCurrentEventNumber() == scenario.getTotalNumberOfEvents()) {
			// simulation completed; optional methods to be defined
		} else {
			collectData(scenario);
			performLBT(scenario);
			updateIRSS(scenario);
		}
	}

	/**
	 * initializes the parameter used by the PPP
	 * 
	 * @param scenario
	 */
	private void initPlugin(ScenarioInfo scenario) {
		devicesActive = new Vector<DataCollectedForWorkshop>();
		devicesInactive = new Vector<DataCollectedForWorkshop>();
		transmitPower = scenario.getInterferingLink(1).getTransmitter().getSuppliedPower();
		antennaPattern = scenario.getInterferingLink(1).getTransmitter().getSphericalAntennaPattern();
		localFreeSpace = new FreeSpaceModel();
		frequencyLBT = scenario.getInterferingLink(1).getFrequency();
	}

	/**
	 * collects the data of each device and puts the device to the list "active"
	 * 
	 * @param scenario
	 */
	private void collectData(ScenarioInfo scenario) {
		List<InterferingLink> interferingLinks = scenario.getInterferingLinks();
		devicesActive = new Vector<DataCollectedForWorkshop>();
		devicesInactive = new Vector<DataCollectedForWorkshop>();
		for (InterferingLink ilk : interferingLinks) {
			device = new DataCollectedForWorkshop();
			device.event = scenario.getCurrentEventNumber();
			device.id = ilk.getLinkIndex();
			device.originLink = ilk.getOrigLinkIndex(); // if needed to distinguish different interfering links
			device.transmitter = ilk.getTransmitter();
			device.receiver = ilk.getReceiver();
			device.iRSSunw = ilk.getIRSSUnwanted();
			device.iRSSblock = ilk.getIRSSBlocking();
			device.active = 1; // set to active
			devicesActive.add(device); // adds this device to the list 'active'
		}
	}

	/**
	 * performs the LBT feature between all pairs of devices <br/>
	 * and puts the switched off devices to the list "inactive"
	 * 
	 * @param scenario
	 */
	private void performLBT(ScenarioInfo scenario) {
		DataCollectedForWorkshop deviceListen = new DataCollectedForWorkshop();
		DataCollectedForWorkshop deviceToDetect = new DataCollectedForWorkshop();
		for (int i = 0; i < devicesActive.size() - 1; i++) {
			deviceToDetect = devicesActive.get(i); // copies the device
			for (int j = i + 1; j < devicesActive.size(); j++) {
				deviceListen = devicesActive.get(j); // copies the device
				double powerReceived = calculatePowerReceived(deviceListen, deviceToDetect);
				if (powerReceived > threshold) {
					// switch deviceListen off, add it to the list 'inactive' and remove it from the list 'active'
					deviceListen.active = 0;
					devicesInactive.add(deviceListen);
					devicesActive.remove(deviceListen); // reduces also the size of devicesActive
				}
			}
		}

	}

	/**
	 * 
	 * @param deviceListen
	 * @param deviceToDetect
	 * @return by the deviceListen received power
	 */
	private double calculatePowerReceived(DataCollectedForWorkshop deviceListen, DataCollectedForWorkshop deviceToDetect) {
		double pathLoss = 0;
		double distance = deviceListen.transmitter.getDistanceTo(deviceToDetect.transmitter);
		localFreeSpace.setVariationsSelected(false);
		try {
			pathLoss = localFreeSpace.evaluate(frequencyLBT, distance, deviceListen.transmitter.getAntennaHeight(),
					deviceToDetect.transmitter.getAntennaHeight());
		} catch (ModelException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		double antennaGain = getAntennaGain(deviceListen, deviceToDetect, distance);
		double powerReceived = deviceToDetect.transmitter.getSuppliedPower() - pathLoss
				+ deviceToDetect.transmitter.getAntennaGain() + antennaGain;

		return powerReceived;
	}

	/**
	 * calculates the resulting antenna gain considering both antennas
	 * 
	 * @param deviceListen
	 * @param deviceToDetect
	 * @param distance
	 * @param deltaY
	 * @param deltaX
	 * @return
	 */
	private double getAntennaGain(DataCollectedForWorkshop deviceListen, DataCollectedForWorkshop deviceToDetect,
			double distance) {
		double deltaX = deviceListen.transmitter.getPositionX() - deviceToDetect.transmitter.getPositionX();
		double deltaY = deviceListen.transmitter.getPositionY() - deviceToDetect.transmitter.getPositionY();
		double azimuthListenToDetect = getAngle(deltaX, deltaY); //  considering only the locations 
		double azimuthDetectToListen = 360 - azimuthListenToDetect;

		double azimuthListenRx = getAngle(
				// angle between ILT and its ILR
				deviceListen.receiver.getPositionX() - deviceListen.transmitter.getPositionX(),
				deviceListen.receiver.getPositionY() - deviceListen.transmitter.getPositionY());

		double azimuthDetectRx = getAngle(
				deviceToDetect.receiver.getPositionX() - deviceToDetect.transmitter.getPositionX(),
				deviceToDetect.receiver.getPositionY() - deviceToDetect.transmitter.getPositionY());

		double azimuthListenToDetectResulting = azimuthListenRx - azimuthListenToDetect; // considers the resulting angle for the antenna pattern
		if (azimuthListenToDetectResulting < 0)
			azimuthListenToDetectResulting += 360;
		if (azimuthListenToDetectResulting > 180)
			azimuthListenToDetectResulting = 360 - azimuthListenToDetectResulting; // due to spherical antenna

		double azimuthDetectToListenResulting = azimuthDetectRx - azimuthDetectToListen; // considers the resulting angle for the antenna pattern
		if (azimuthDetectToListenResulting < 0)
			azimuthDetectToListenResulting += 360;
		if (azimuthDetectToListenResulting > 180)
			azimuthDetectToListenResulting = 360 - azimuthDetectToListenResulting; // due to spherical antenna

		double elevationListenRx = getAngle(deviceListen.transmitter.getDistanceTo(deviceListen.receiver) * 1000,
				deviceListen.receiver.getAntennaHeight() - deviceListen.transmitter.getAntennaHeight()) - 360;

		double elevationDetectRX = getAngle(deviceToDetect.transmitter.getDistanceTo(deviceToDetect.receiver) * 1000,
				deviceToDetect.receiver.getAntennaHeight() - deviceToDetect.transmitter.getAntennaHeight()) - 360;

		double elevationListenToDetect = getAngle(
				deviceListen.transmitter.getDistanceTo(deviceToDetect.transmitter) * 1000,
				deviceToDetect.transmitter.getAntennaHeight() - deviceListen.transmitter.getAntennaHeight());

		/* for the purpose of the presentation not continued
		 * the impact of the elevation angles as well as the angles between ILT and ILR are ignored 
		 * the elevation angle is for the calculation set to zero
		 */

		double rElevation = 0;

		double gainListenToDetect = calculateGain(rElevation, azimuthListenToDetect);
		double gainDetectToListen = calculateGain(rElevation, azimuthDetectToListen);

		return gainListenToDetect + gainDetectToListen;
	}

	/**
	 * method supporting getAntennaGain
	 * 
	 * @param azimuthListenToDetect
	 * @param rElevation
	 * 
	 * @return
	 */
	private double calculateGain(double rElevation, double azimuthListenToDetect) {
		double antennaGain = 0;
		try {
			double sphericalAngle = Mathematics.acosD(Mathematics.cosD(rElevation)
					* Mathematics.cosD(azimuthListenToDetect));
			antennaGain = antennaPattern.evaluate(sphericalAngle);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return antennaGain;
	}

	/**
	 * calculates the angle
	 * 
	 * @param deltaX
	 * @param deltaY
	 * @return angle
	 */
	private double getAngle(double deltaX, double deltaY) {
		double angle = Mathematics.atan2D(deltaX, deltaY);
		if (angle < 0)
			angle += 360;
		return angle;
	}

	/**
	 * decreases the origin values of iRSS by 200 dB for the devices on the list
	 * "inactive"
	 * 
	 * @param scenario
	 */
	private void updateIRSS(ScenarioInfo scenario) {
		double sumUnw = 0, sumBlock = 0;
		int countLink = 1;
		for (int i = 0; i < devicesInactive.size(); i++) { // switched off devices
			devicesInactive.get(i).iRSSunw -= 200;
			devicesInactive.get(i).iRSSblock -= 200;
			sumUnw += Math.pow(10, devicesInactive.get(i).iRSSunw / 10);
			sumBlock += Math.pow(10, devicesInactive.get(i).iRSSblock / 10);
			countLink = Math.max(countLink, devicesInactive.get(i).originLink); // considers the highest number
		}

		for (int j = 0; j < devicesActive.size() - 1; j++) {// remaining active devices
			sumUnw += Math.pow(10, devicesActive.get(j).iRSSunw / 10);
			sumBlock += Math.pow(10, devicesActive.get(j).iRSSblock / 10);
			countLink = Math.max(countLink, devicesActive.get(j).originLink); // considers the highest number
		}

		double iRRSunwCorrected = 10 * Math.log10(sumUnw);
		double iRSSblockCorrected = 10 * Math.log10(sumBlock);

		for (int k = 1; k <= countLink; k++) {
			scenario.getProcessingResults().setIRSSUnwanted(k, iRRSunwCorrected);
			scenario.getProcessingResults().setIRSSBlocking(k, iRSSblockCorrected);
		}

	}

	/* end of methods added */

	public int getNumberOfParameters() {
		return 1;
	}

	public ParameterType getParameterType(int id) {
		switch (id) {
		case 1: {
			return ParameterType.Double;
		}

		default: {
			throw new IllegalArgumentException("illegal parameter id");
		}

		}
	}

	public void setParameterValue(int id, Object value) {
		try {
			switch (id) {
			case 1: {
				threshold = (Double) value;
				break;
			}
			default: {
				throw new IllegalArgumentException("illegal parameter id");
			}
			}
		} catch (Exception e) {

		}
	}

	public Object getParameterValue(int id) {
		switch (id) {
		case 1: {
			return threshold;
		}
		default: {
			throw new IllegalArgumentException("illegal parameter id");
		}

		}
	}

	public String getParameterName(int id) {
		switch (id) {
		case 1: {
			return "sensing threshold [dBm]";
		}

		default: {
			throw new IllegalArgumentException("illegal parameter id");
		}

		}
	}

	/*
	 * check consistency of the workspace
	 */
	public boolean check(ScenarioInfo info) {
		checkPropagationModel = true;

		String warning = "<html>";
		warning += "</html>";

		if (checkPropagationModel == false) {
			info.addConsistencyWarning(warning);
			return false;
		} else {
			return true;
		}
	}

	@Override
	public Vector<String> getAllowedParameterValues(int index) {
		// TODO Auto-generated method stub
		return null;
	}

}

class DataCollectedForWorkshop {
	int id, originLink, event;
	int active; // 1 = active   0 = switched off	
	double iRSSunw, iRSSblock;
	InterferingTransmitter transmitter;
	Transceiver receiver;

}
